
#ifndef ETH_DESCR_H_
#define ETH_DESCR_H_


/** \brief Convert local DSPR address to global DSPR address which can be accessed from the SRI bus.
 * Use this macro to convert a local DSPR address (in segment 0xd00.....) to
 * a global DSPR address (in segment 0x700....., 0x600....., 0x500..... downwards) depending on
 * the CPU number.
 * Example usage:
 *  \code
 *     dmaChConfig.sourceAddress      = IFXCPU_GLB_ADDR_DSPR(IfxCpu_getCoreId(), &sourceBuffer[i][0]);
 *     dmaChConfig.destinationAddress = IFXCPU_GLB_ADDR_DSPR(IfxCpu_getCoreId(), &destinationBuffer[i][0]);
 *  \endcode
 */
#define IFXCPU_GLB_ADDR_DSPR(cpu, address) ((((((unsigned)(address) & 0xF0000000) == 0xD0000000) ? ((((unsigned)(address) & 0x000fffff) | 0x70000000) - ((cpu) * 0x10000000)) : (unsigned)(address))))

/** \brief Convert local PSPR address to global PSPR address which can be accessed from the SRI bus.
 * Use this macro to convert a local PSPR address (in segment 0xc......) to
 * a global PSPR address (in segment 0x701....., 0x601....., 0x501..... downwards) depending on
 * the CPU number.
 *
 *   Example usage:
 *   \code
 *     dmaChConfig.sourceAddress      = IFXCPU_GLB_ADDR_PSPR(IfxCpu_getCoreId(), &sourceBufferInPsprMemory);
 *     dmaChConfig.destinationAddress = IFXCPU_GLB_ADDR_PSPR(IfxCpu_getCoreId(), &destinationBufferInPsprMemory);
 *   \endcode
 */
#define IFXCPU_GLB_ADDR_PSPR(cpu, address) ((((unsigned)(address) & 0x000fffff) | 0x70100000) - ((cpu) * 0x10000000))

typedef struct
{
    uint32 ext : 1;    /**< \brief Extended Status Available/Rx MAC Address */
    uint32 CE : 1;     /**< \brief CRC Error */
    uint32 DBE : 1;    /**< \brief Dribble Bit Error */
    uint32 RE : 1;     /**< \brief Receive Error */
    uint32 RWT : 1;    /**< \brief Receive Watchdog Timeout */
    uint32 FT : 1;     /**< \brief Frame Type */
    uint32 LC : 1;     /**< \brief Late Collision */
    uint32 IPC : 1;    /**< \brief IPC Checksum Error/Giant Frame */
    uint32 LS : 1;     /**< \brief Last Descriptor */
    uint32 FS : 1;     /**< \brief First Descriptor */
    uint32 VLAN : 1;   /**< \brief VLAN Tag */
    uint32 OE : 1;     /**< \brief Overflow Error */
    uint32 LE : 1;     /**< \brief Length Error */
    uint32 SAF : 1;    /**< \brief Source Address Filter Fail */
    uint32 DE : 1;     /**< \brief Descriptor Error */
    uint32 ES : 1;     /**< \brief Error Summary, ES = PCE | CE | RE | RWT | LC | IPC | OE | DE */
    uint32 FL : 14;    /**< \brief Frame Length */
    uint32 AFM : 1;    /**< \brief Destination Address Filter Fail */
    uint32 OWN : 1;    /**< \brief Own Bit, 1 = own by DMA */
} IfxEth_AltRxDescr0_Bits;

typedef struct
{
    uint32 RBS1 : 13;   /**< \brief Receive Buffer 1 Size */
    uint32 resv1 : 1;   /**< \brief reserved */
    uint32 RCH : 1;     /**< \brief Second Address Chained */
    uint32 RER : 1;     /**< \brief Receive End of Ring */
    uint32 RBS2 : 13;   /**< \brief Receive Buffer 2 Size */
    uint32 resv : 2;    /**< \brief reserved) */
    uint32 DIC : 1;     /**< \brief Disable Interrupt on Completion */
} IfxEth_AltRxDescr1_Bits;

typedef struct
{
    uint32 DB : 1;      /**< \brief Deferred bit */
    uint32 UF : 1;      /**< \brief Underflow error */
    uint32 ED : 1;      /**< \brief Excessive deferral */
    uint32 CC : 4;      /**< \brief Collision count */
    uint32 VLAN : 1;    /**< \brief VLAN TAG */
    uint32 EC : 1;      /**< \brief Excessive Collision */
    uint32 LC : 1;      /**< \brief Late Collision */
    uint32 NC : 1;      /**< \brief No Carrier */
    uint32 LOC : 1;     /**< \brief Loss of Carrier */
    uint32 PCE : 1;     /**< \brief Payload Checksum Error */
    uint32 FF : 1;      /**< \brief Frame Flushed */
    uint32 JT : 1;      /**< \brief Jabber Timeout */
    uint32 ES : 1;      /**< \brief Error Summary, ES = JT | FF | LOC | NC | LC | EC | ED | UF */
    uint32 IHE : 1;     /**< \brief IP Header Error */
    uint32 TTSS : 1;    /**< \brief Transmit Time Stamp Status */
    uint32 resv : 2;    /**< \brief (reserved) */
    uint32 TCH : 1;     /**< \brief Second Address Chained */
    uint32 TER : 1;     /**< \brief Transmit End of Ring */
    uint32 CIC : 2;     /**< \brief Checksum Insertion Control */
    uint32 resv1 : 1;   /**< \brief (Reserved) */
    uint32 TTSE : 1;    /**< \brief Transmit Time Stamp Enable */
    uint32 DP : 1;      /**< \brief Disable Padding */
    uint32 DC : 1;      /**< \brief Disable CRC */
    uint32 FS : 1;      /**< \brief First Segment */
    uint32 LS : 1;      /**< \brief Last Segment */
    uint32 IC : 1;      /**< \brief Interrupt on Completion */
    uint32 OWN : 1;     /**< \brief Own Bit, 1 = own by DMA */
} IfxEth_AltTxDescr0_Bits;

typedef struct
{
    uint32 TBS1 : 13;   /**< \brief Transmit Buffer 1 Size */
    uint32 resv1 : 3;   /**< \brief (reserved) */
    uint32 TBS2 : 13;   /**< \brief Transmit Buffer 2 Size */
    uint32 resv2 : 3;   /**< \brief (reserved) */
} IfxEth_AltTxDescr1_Bits;

typedef union
{
    IfxEth_AltRxDescr0_Bits A;     /**< \brief Structure for RX descriptor DWORD 0 Bit field access */
    uint32                  U;     /**< \brief Unsigned long access */
} IfxEth_RxDescr0;

typedef union
{
    IfxEth_AltRxDescr1_Bits A;     /**< \brief Structure for RX descriptor DWORD 1 Bit field access */
    uint32                  U;     /**< \brief unsigned long access */
} IfxEth_RxDescr1;

typedef union
{
    uint32 U;     /**< \brief unsigned long access */
} IfxEth_RxDescr2;

typedef union
{
    uint32 U;     /**< \brief unsigned long access */
} IfxEth_RxDescr3;

typedef union
{
    IfxEth_AltTxDescr0_Bits A;     /**< \brief Structure for TX descriptor DWORD 0 Bit field access */
    uint32                  U;     /**< \brief Unsigned long access */
} IfxEth_TxDescr0;

typedef union
{
    IfxEth_AltTxDescr1_Bits A;     /**< \brief Structure for RX descriptor DWORD 1 Bit field access */
    uint32                  U;     /**< \brief unsigned long access */
} IfxEth_TxDescr1;

typedef union
{
    uint32 U;     /**< \brief unsigned long access */
} IfxEth_TxDescr2;

typedef union
{
    uint32 U;     /**< \brief unsigned long access */
} IfxEth_TxDescr3;

typedef struct
{
    IfxEth_RxDescr0 RDES0;     /**< \brief RX descriptor DWORD 0 */
    IfxEth_RxDescr1 RDES1;     /**< \brief RX descriptor DWORD 1 */
    IfxEth_RxDescr2 RDES2;     /**< \brief RX descriptor DWORD 2 */
    IfxEth_RxDescr3 RDES3;     /**< \brief RX descriptor DWORD 3 */
} IfxEth_RxDescr;

typedef struct
{
    IfxEth_TxDescr0 TDES0;     /**< \brief TX descriptor DWORD 0 */
    IfxEth_TxDescr1 TDES1;     /**< \brief TX descriptor DWORD 1 */
    IfxEth_TxDescr2 TDES2;     /**< \brief TX descriptor DWORD 2 */
    IfxEth_TxDescr3 TDES3;     /**< \brief TX descriptor DWORD 3 */
} IfxEth_TxDescr;


typedef union
{
    IfxEth_RxDescr items[ETH_RX_BUFFERS];
    uint32         U[ETH_RX_BUFFERS][ETH_DESCR_SIZE];
} IfxEth_RxDescrList;

typedef union
{
    IfxEth_TxDescr items[ETH_TX_BUFFERS];
    uint32         U[ETH_TX_BUFFERS][ETH_DESCR_SIZE];
} IfxEth_TxDescrList;


typedef struct
{
    uint32 PCE : 1;    /**< \brief Rx MAC Address/Payload Checksum Error */
    uint32 CE : 1;     /**< \brief CRC Error */
    uint32 DBE : 1;    /**< \brief Dribble Bit Error */
    uint32 RE : 1;     /**< \brief Receive Error */
    uint32 RWT : 1;    /**< \brief Receive Watchdog Timeout */
    uint32 FT : 1;     /**< \brief Frame Type */
    uint32 LC : 1;     /**< \brief Late Collision */
    uint32 IPC : 1;    /**< \brief IPC Checksum Error/Giant Frame */
    uint32 LS : 1;     /**< \brief Last Descriptor */
    uint32 FS : 1;     /**< \brief First Descriptor */
    uint32 VLAN : 1;   /**< \brief VLAN Tag */
    uint32 OE : 1;     /**< \brief Overflow Error */
    uint32 LE : 1;     /**< \brief Length Error */
    uint32 SAF : 1;    /**< \brief Source Address Filter Fail */
    uint32 DE : 1;     /**< \brief Descriptor Error */
    uint32 ES : 1;     /**< \brief Error Summary, ES = PCE | CE | RE | RWT | LC | IPC | OE | DE */
    uint32 FL : 14;    /**< \brief Frame Length */
    uint32 AFM : 1;    /**< \brief Destination Address Filter Fail */
    uint32 OWN : 1;    /**< \brief Own Bit, 1 = own by DMA */
} IfxEth_RxDescr0_Bits;

typedef struct
{
    uint32 RBS1 : 11;   /**< \brief Receive Buffer 1 Size */
    uint32 RBS2 : 11;   /**< \brief Receive Buffer 2 Size */
    uint32 resv : 2;    /**< \brief (reserved) */
    uint32 RCH : 1;     /**< \brief Second Address Chained */
    uint32 RER : 1;     /**< \brief Receive End of Ring */
    uint32 resv2 : 5;   /**< \brief (reserved) */
    uint32 DIC : 1;     /**< \brief Disable Interrupt on Completion */
} IfxEth_RxDescr1_Bits;

typedef struct
{
    uint32 DB : 1;      /**< \brief Deferred Bit */
    uint32 UF : 1;      /**< \brief Underflow Error */
    uint32 ED : 1;      /**< \brief Excessive Deferral */
    uint32 CC : 4;      /**< \brief Collision Count */
    uint32 VLAN : 1;    /**< \brief VLAN Tag */
    uint32 EC : 1;      /**< \brief Excessive Collision */
    uint32 LC : 1;      /**< \brief Late Collision */
    uint32 NC : 1;      /**< \brief No Carrier */
    uint32 LOC : 1;     /**< \brief Loss of Carrier */
    uint32 PCE : 1;     /**< \brief Payload Checksum Error */
    uint32 FF : 1;      /**< \brief Frame Flushed */
    uint32 JT : 1;      /**< \brief Jabber Timeout */
    uint32 ES : 1;      /**< \brief Error Summary, ES = JT | FF | LOC | NC | LC | EC | ED | UF */
    uint32 IHE : 1;     /**< \brief IP Header Error */
    uint32 TTSS : 1;    /**< \brief Tx Time Stamp Status */
    uint32 resv : 13;   /**< \brief (reserved) */
    uint32 OWN : 1;     /**< \brief Own Bit, 1 = own by DMA */
} IfxEth_TxDescr0_Bits;

typedef struct
{
    uint32 TBS1 : 11;   /**< \brief Transmit Buffer 1 Size */
    uint32 TBS2 : 11;   /**< \brief Transmit Buffer 2 Size */
    uint32 TTSE : 1;    /**< \brief Transmit Time Stamp Enable */
    uint32 DP : 1;      /**< \brief Disable Padding */
    uint32 TCH : 1;     /**< \brief Second Address Chained */
    uint32 TER : 1;     /**< \brief Transmit End of Ring */
    uint32 DC : 1;      /**< \brief Disable CRC */
    uint32 CIC : 2;     /**< \brief Checksum Insertion Control */
    uint32 FS : 1;      /**< \brief First Segment */
    uint32 LS : 1;      /**< \brief Last Segment */
    uint32 IC : 1;      /**< \brief Interrupt on Completion */
} IfxEth_TxDescr1_Bits;

#endif /* ETH_DESCR_H_ */
