/*
 * TFT.c
 *
 *  Created on: 21.05.2015
 *      Author: AllanTengg
 */

#include "tc27xc/Ifx_reg.h"
#include <machine/intrinsics.h>
#include <machine/wdtcon.h>
#include <stdlib.h>
#include <string.h>
#include "led.h"
#include "debug.h"
#include "types.h"
#include "TFT.h"

static Ifx_QSPI_BACON s_bacon0;

volatile uint32 tft_status = 0;
volatile uint16 tft_id = 0;

#define TFT_ChipSelect();    P20_OMCR.B.PCL6 = 1;
#define TFT_ChipUnselect();  P20_OMSR.B.PS6 = 1;

uint16 Pixel_Buff[8*TFT_XSIZE];  //to have the correct 4-byte alignment

const uint16 CBM_Palette[16] =
{
   COLOR_BLACK,
   COLOR_WHITE,
   COLOR_RED,
   COLOR_CYAN,
   COLOR_PURPLE,
   COLOR_GREEN,
   COLOR_BLUE,
   COLOR_YELLOW,
   COLOR_ORANGE,
   COLOR_BROWN,
   COLOR_LIGHTRED,
   COLOR_DARKGRAY,
   COLOR_MEDGRAY,
   COLOR_LIGHTGREEN,
   COLOR_LIGHTBLUE,
   COLOR_LIGHTGRAY
};

const uint16 VGA_Palette[16] =
{
	RGB(0,0,0),
	RGB(255,0,0),
	RGB(0,255,0),
	RGB(255,255,0),
	RGB(0,0,255),
	RGB(255,0,255),
	RGB(0,255,255),
	RGB(128,128,128),
	RGB(64,64,64),
	RGB(255,128,0),
	RGB(128,255,128),
	RGB(255,255,128),
	RGB(128,128,255),
	RGB(255,128,255),
	RGB(128,255,255),
	RGB(255,255,255)
};

unsigned int QSPI0_Transfer(unsigned int v, unsigned char write_bacon)
{
	unsigned int tmp;
	static unsigned int rx;
	static unsigned int stat;

	while (QSPI0_STATUS.B.RXFIFOLEVEL > 0)
	  tmp = QSPI0_RXEXIT.U;

	if (write_bacon)
	{
	   QSPI0_GLOBALCON.B.EN = 0;     //request halt state
	   QSPI0_BACONENTRY.U = s_bacon0.U;
	   QSPI0_DATAENTRY0.U = v;
	   QSPI0_GLOBALCON.B.EN = 1;     //request run state
	}
	else
	{
	   QSPI0_DATAENTRY0.U = v;
	}

	while (QSPI0_STATUS.B.TXFIFOLEVEL > 0);  //wait until fifo is empty
	while (QSPI0_STATUS.B.RXFIFOLEVEL < 2);  //wait until values are available for CPU

	rx = QSPI0_RXEXIT.U;
	stat = QSPI0_RXEXIT.U;

	return rx;
}

void QSPI0_Init()
{
	//Configure the backlight pin to output and enable it
	P20_IOCR12.B.PC13 = 0x10;
	P20_OMSR.B.PS13 = 1;

	//Use QSPI0 for TFT communication (I control the CS line myself!)
	P20_OMSR.B.PS6 = 1;            //Begin with inactive CS line
	P20_IOCR4.B.PC6 = 0x10;        //Conf. P20.6 for port function (CS)
	P20_IOCR8.B.PC11 = 0x13;       //Conf. P20.11 for SCLK0 function
	P20_IOCR12.B.PC12 = 0x00;      //Conf. P20.12 for MRST0 function
	P20_IOCR12.B.PC14 = 0x13;      //Conf. P20.14 for MTSR0 function

	//IMPORTANT: Need to set faster port driver to support full SPI frequency
	unlock_wdtcon();
	P20_PDR0.B.PD6 = 0x01;
	P20_PDR1.B.PD11 = 0x01;
	P20_PDR1.B.PD14 = 0x01;

	//Enable QSPI0 module by clearing its disable bit
	QSPI0_CLC.U = 0;
	while (QSPI0_CLC.U & 2);    //wait for confirmation
	lock_wdtcon();

	//Route signals from/to QSPI0 module
	QSPI0_PISEL.B.MRIS = 0x00;    //use MRST0A pin
	//QSPI0_SSOC.B.OEN = (1<<8);    //Enable SLSO8

	//Configure QSPI0 globally, SPI->Parallel converter max. clock ^= 50 MHz
	QSPI0_GLOBALCON.B.TQ = 1;     //use fBaud2/2 for time quantum (= 1 / 90 MHz = 11.1 ns)
	QSPI0_GLOBALCON.B.MS = 0;     //we are Master on this SPI bus
	QSPI0_GLOBALCON.B.SI = 1;     //allow status word injection after every transmission (BACON|DATAWORD -> DATAWORD|STATUS)
    QSPI0_GLOBALCON1.B.TXEN = 0;  //enable/disable TX interrupt
    QSPI0_GLOBALCON1.B.RXEN = 0;  //enable/disable RX interrupt
    QSPI0_GLOBALCON1.B.TXFM = 1;  //single move mode for RX and TX
    QSPI0_GLOBALCON1.B.RXFM = 1;
    QSPI0_GLOBALCON1.B.PT1 = 2;   //set PT1 with SOF event
    QSPI0_GLOBALCON.B.EN = 1;     //request run state

    //TODO: Clock Timing optimieren...

    //Configure channel SLSO8 via ECON0
    QSPI0_ECON0.B.A = 1;          //2 Tq before sample
    QSPI0_ECON0.B.B = 1;          //1 Tq delay in sampling
    QSPI0_ECON0.B.C = 1;          //2 Tq after sample
    QSPI0_ECON0.B.Q = 0;          //(Q+1)*(A+1+B+C) >=4
    QSPI0_ECON0.B.CPH = 0;        //5 * Tq
    QSPI0_ECON0.B.CPOL = 0;       //Idle = Low, Sample with rising edge of CLK
    QSPI0_ECON0.B.PAREN = 0;      //No parity checks needed

    //Configure the bacon shadow register
    s_bacon0.B.LAST = 1;          //Configure for single transfer
    s_bacon0.B.IPRE = 0;          //Prescaler for delay = 1
    s_bacon0.B.IDLE = 0;          //1 Tq should be enough
    s_bacon0.B.LPRE = 1;          //Prescaler for leading delay = 4
    s_bacon0.B.LEAD = 1;          //8 Tq should be enough lead delay
    s_bacon0.B.TPRE = 1;          //Prescaler for trailing delay = 4
    s_bacon0.B.TRAIL= 1;          //4 Tq should be enough
    s_bacon0.B.PARTYP=0;          //Parity is not used
    s_bacon0.B.UINT = 0;          //For now do not use interrupt
    s_bacon0.B.MSB =  1;          //MSB first
    s_bacon0.B.BYTE = 0;          //Datalength is specified in bits
    s_bacon0.B.DL =  31;          //32 bits per frame
    s_bacon0.B.CS =   8;          //Use SLSO0.8
}

void TFT_BacklightOnOff (unsigned char st)
{
	if (st)
		P20_OMSR.B.PS13 = 1;
	else
		P20_OMCR.B.PCL13 = 1;
}

void TFT_Write_Data (uint8 regaddr, uint16 data)
{
	TFT_ChipSelect();
    QSPI0_Transfer(((unsigned)regaddr << 22) | ((unsigned)data << 6), 1);
    TFT_ChipUnselect();
}

uint32 TFT_Read_Data (uint8 regaddr)
{
    uint32 tx_data[2];
    uint32 rx_data[2];
    uint16 cnt;

    for (cnt=0; cnt<2; cnt++)
    {
    	tx_data[cnt] = 0;
    	rx_data[cnt] = 0;
    }

	TFT_ChipSelect();

    // We need 16 bit for address and 26 bits for value = 42 bits total
    // thus we change the datawidth to 21 bit
    s_bacon0.B.DL = 20;

    s_bacon0.B.LAST = 0;
    rx_data[0] = QSPI0_Transfer(tx_data[0], 1);

    s_bacon0.B.LAST = 1;
    rx_data[1] = QSPI0_Transfer(tx_data[1], 1);

    // set back to 32 bit transfer
    s_bacon0.B.DL = 31;

    TFT_ChipUnselect();

    return rx_data[1];
}

void TFT_Read_Data_ili9341 (uint8 regaddr, uint16 *puiData, uint32 count)
{
	uint32 i;
    uint16 tx_data;

	TFT_ChipSelect();

    // we write first the address (10 bit) => we change the datawidth to 10 bit
    s_bacon0.B.DL = 9;

    // calculate the command and address value
    tx_data = (0x03 << 8) | regaddr;

    // and send the address to ILI9341
    s_bacon0.B.LAST = 0;
    QSPI0_Transfer(tx_data, 1);

    // our values haven 16 bit  => we change the datawidth to 16 bit
    s_bacon0.B.DL = 15;

    // send dummy values and read the values from ILI9341
    for (i=0; i<count-1; i++)
    {
    	if (i==0)
    		puiData[i] = QSPI0_Transfer(0, 1);
    	else
    	    puiData[i] = QSPI0_Transfer(0, 0);
    }

    s_bacon0.B.LAST = 1;
	puiData[count-1] = QSPI0_Transfer(0, 1);

	TFT_ChipUnselect();

    // we terminate the transfer by generate one SCLK without transfer
    // we make sure that the SCLK will be low when we switch to general output
    P20_OMCR.B.PCL11 = 1;
    P20_IOCR8.B.PC11 = 0x10;

    // toggle the SCLK
    P20_OMSR.B.PS11 = 1;
    _nop(); _nop(); _nop(); _nop();
    P20_OMCR.B.PCL11 = 1;

    // set the SCLK back to used alternate output
	P20_IOCR8.B.PC11 = 0x13;

    // Finally set back to 32 bit transfer
    s_bacon0.B.DL = 31;
}

void TFT_Write_Data_ili9341 (uint8 regaddr, uint16 *puiData, uint32 count)
{
	uint32 i;
    uint16 tx_data;

	TFT_ChipSelect();

    // we write first the address (10 bit) => we change the datawidth to 10 bit
    s_bacon0.B.DL = 9;

    // calculate the command and address value
    tx_data = (1 << 8) | regaddr;

    // send the address to ILI9341
    s_bacon0.B.LAST = 0;
    QSPI0_Transfer(tx_data, 1);

    // our values haven 16 bit => we change the datawidth to 16 bit
    s_bacon0.B.DL = 15;

    // send the values to ILI9341
    for (i=0; i<count-1; i++)
    {
    	if (i==0)
    	   QSPI0_Transfer(puiData[i], 1);
    	else
    	   QSPI0_Transfer(puiData[i], 0);
    }

    s_bacon0.B.LAST = 1;
    QSPI0_Transfer(puiData[count-1], 1);

	TFT_ChipUnselect();

    // we terminate the transfer by generate one SCLK without transfer
    // we make sure that the SCLK will be low when we switch to general output
    P20_OMCR.B.PCL11 = 1;
    P20_IOCR8.B.PC11 = 0x10;

    // toggle the SCLK
    P20_OMSR.B.PS11 = 1;
    _nop(); _nop(); _nop(); _nop();
    P20_OMCR.B.PCL11 = 1;

    // set the SCLK back to used alternate output
	P20_IOCR8.B.PC11 = 0x13;

    // Finally set back to 32 bit transfer
    s_bacon0.B.DL = 31;
}

static void delay_us (uint32 time)
{
    volatile uint32 test;
    uint32 i;
    test = 0;
    for (i = 0; i < time; i += 1)
    {
        test += 1;
    }
}

static void delay_ms (uint32 time)
{
    while (time--)
    {
        delay_us (1000);
    }
}


void TFT_Init()
{
	uint16 uwData[16];

	QSPI0_Init();

    // To force the SPI-Parallel converter we need to send a single SCLK pulse
    // we make sure that the SCLK will be low when we switch to general output
    P20_OMCR.B.PCL11 = 1;
    P20_IOCR8.B.PC11 = 0x10;

    // toggle the SCLK
    P20_OMSR.B.PS11 = 1;
    _nop(); _nop(); _nop(); _nop();
    P20_OMCR.B.PCL11 = 1;

    // set the SCLK back to used alternate output
	P20_IOCR8.B.PC11 = 0x13;


	//Now detect the chip that is connected
    tft_id = 0;
    tft_id = TFT_Read_Data (0);

    if (tft_id == 0)
    {
        // check for ILI9341 controller
        TFT_Read_Data_ili9341 (0xD3, &uwData[0], 5);

        //Sometimes the first sequence fails (prop. indeterministic state)
        TFT_Read_Data_ili9341 (0xD3, &uwData[0], 5);

        if ((uwData[3] == 0x0093) && (uwData[4] == 0x0041))
        {
            tft_id = 0x9341;
        }
        else
        {
            for (;;);
        }
    }
    else
    {
        for (;;);
    }

	/************* Start Initial Sequence **********/
	uwData[0] = 0x0000;
	uwData[1] = 0x0083;
	uwData[2] = 0x0030;
	uwData[3] = 0x0000;
	TFT_Write_Data_ili9341(0xCF, &uwData[0], 4); // setting from display supplier

	uwData[0] = 0x0064;
	uwData[1] = 0x0003;
	uwData[2] = 0x0012;
	uwData[3] = 0x0081;
	uwData[4] = 0x0000;
	TFT_Write_Data_ili9341(0xED, &uwData[0], 5); // setting from display supplier

	uwData[0] = 0x0085;
	uwData[1] = 0x0000;
	uwData[2] = 0x0078;
	uwData[3] = 0x0000;
	TFT_Write_Data_ili9341(0xE8, &uwData[0], 4); // setting from display supplier

	uwData[0] = 0x0039;
	uwData[1] = 0x002C;
	uwData[2] = 0x0000;
	uwData[3] = 0x0034;
	uwData[4] = 0x0002;
	uwData[5] = 0x0000;
	TFT_Write_Data_ili9341(0xCB, &uwData[0], 6); // setting from display supplier

	uwData[0] = 0x0020;
	uwData[1] = 0x0000;
	TFT_Write_Data_ili9341(0xF7, &uwData[0], 2); // setting from display supplier

	uwData[0] = 0x0000;
	uwData[1] = 0x0000;
	uwData[2] = 0x0000;
	TFT_Write_Data_ili9341(0xEA, &uwData[0], 3); // setting from display supplier

	uwData[0] = 0x0019;  // VRH[5:0]
	uwData[1] = 0x0000;
	TFT_Write_Data_ili9341(0xC0, &uwData[0], 2); // Power Control 1

	uwData[0] = 0x0011;  // SAP[2:0];BT[3:0]
	uwData[1] = 0x0000;
	TFT_Write_Data_ili9341(0xC1, &uwData[0], 2); // Power Control 2

	uwData[0] = 0x0031;
	uwData[1] = 0x003C;
	uwData[2] = 0x0000;
	TFT_Write_Data_ili9341(0xC5, &uwData[0], 3); // VCM Control 1

	uwData[0] = 0x00B0;
	uwData[1] = 0x0000;
	TFT_Write_Data_ili9341(0xC7, &uwData[0], 2); // VCM Control 2

	uwData[0] = 0x0028;  // MV=1; BGR=1
	uwData[1] = 0x0000;
	TFT_Write_Data_ili9341(0x36, &uwData[0], 2); // Memory Access Control

	uwData[0] = 0x0055;  // 16 bit
	uwData[1] = 0x0000;
	TFT_Write_Data_ili9341(0x3A, &uwData[0], 2); // Pixel Format Set

	uwData[0] = 0x0000;
	uwData[1] = 0x0017;
	uwData[2] = 0x0000;
	TFT_Write_Data_ili9341(0xB1, &uwData[0], 3); // Frame Control (in Normal Mode)

	uwData[0] = 0x000A;
	uwData[1] = 0x00A2;
	uwData[2] = 0x0000;
	TFT_Write_Data_ili9341(0xB6, &uwData[0], 3); // Display Function Control

	uwData[0] = 0x0001;
	uwData[1] = 0x0030;
	uwData[2] = 0x0000;
	TFT_Write_Data_ili9341(0xF6, &uwData[0], 3); // Interface Control

	uwData[0] = 0x0000;  // Gamma Function Disable
	uwData[1] = 0x0000;
	TFT_Write_Data_ili9341(0xF2, &uwData[0], 2);  // setting from display supplier

	uwData[0] = 0x0001;     //Gamma curve selected
	uwData[1] = 0x0000;
	TFT_Write_Data_ili9341(0x26, &uwData[0], 2);  // Gamma Set

	uwData[0]  = 0x000F;
	uwData[1]  = 0x0026;
	uwData[2]  = 0x0022;
	uwData[3]  = 0x000A;
	uwData[4]  = 0x0010;
	uwData[5]  = 0x000A;
	uwData[6]  = 0x004C;
	uwData[7]  = 0x00CA;
	uwData[8]  = 0x0036;
	uwData[9]  = 0x0000;
	uwData[10] = 0x0015;
	uwData[11] = 0x0000;
	uwData[12] = 0x0010;
	uwData[13] = 0x0010;
	uwData[14] = 0x0000;
	uwData[15] = 0x0000;
	TFT_Write_Data_ili9341(0xE0, &uwData[0], 16); // Positive Gamma Correction

	uwData[0]  = 0x0000;
	uwData[1]  = 0x0019;
	uwData[2]  = 0x001B;
	uwData[3]  = 0x0005;
	uwData[4]  = 0x000F;
	uwData[5]  = 0x0005;
	uwData[6]  = 0x0033;
	uwData[7]  = 0x0035;
	uwData[8]  = 0x0049;
	uwData[9]  = 0x000F;
	uwData[10] = 0x001F;
	uwData[11] = 0x000F;
	uwData[12] = 0x003F;
	uwData[13] = 0x003F;
	uwData[14] = 0x000F;
	uwData[15] = 0x0000;
	TFT_Write_Data_ili9341(0xE1, &uwData[0], 16); // Negative Gamma Correction

	uwData[0] = 0x0000;
	uwData[1] = 0x0000;
	uwData[2] = ((TFT_XSIZE-1) & 0xFF00)>>8;
	uwData[3] = (TFT_XSIZE-1) & 0x00FF;
	uwData[4] = 0x0000;
	TFT_Write_Data_ili9341(0x2A, &uwData[0], 5);  // Column Address Set

	uwData[0] = 0x0000;
	uwData[1] = 0x0000;
	uwData[2] = ((TFT_YSIZE-1) & 0xFF00)>>8;
	uwData[3] = (TFT_YSIZE-1) & 0x00FF;
	uwData[4] = 0x0000;
	TFT_Write_Data_ili9341(0x2B, &uwData[0], 5);  // Page Address Set

	uwData[0] = 0x0000;
	TFT_Write_Data_ili9341(0x11, &uwData[0], 1);  // Exit Sleep

	delay_ms (120);

	uwData[0] = 0x0000;
	TFT_Write_Data_ili9341(0x29, &uwData[0], 1);  // Display on
}

extern const unsigned char charset[2048];

void TFT_Text(uint16 textcol, uint16 backcol, char *str)
{
	int i;
	int ax, dx, dy;
	unsigned char c, cc, v;

	for (i=0; i<TFT_XSIZE*8; i++)
	   Pixel_Buff[i] = backcol;

	c = *str;
	ax = 0;

	while (c)
	{
		if (c>='a' && c<='z')
		  cc=c-96;
		else
		if (c=='@')
		  cc=0;
		else
		  cc=c;

		for (dy=0; dy<8; dy++)
		{
			v=charset[(cc<<3)+7-dy];
			for (dx=0; dx<8; dx++)
			{
				if (v & 0x80)
					Pixel_Buff[dy*TFT_XSIZE+ax+dx] = textcol;
				else
					Pixel_Buff[dy*TFT_XSIZE+ax+dx] = backcol;

				v <<= 1;
			}
		}
		ax += 8;
		str++;
     	c = *str;
	}

	TFT_Flush_Row_Buff(TFT_XSIZE*8);
}

void TFT_SetXY (uint16 x, uint16 y)
{
	uint16 uwData[3];

	uwData[0] = x >> 8;
	uwData[1] = x;
	uwData[2] = 0;
	TFT_Write_Data_ili9341(0x2A, &uwData[0], 3);  // Column Address Set, we change only the start (2 Parameters)

	uwData[0] = y >> 8;
	uwData[1] = y;
	uwData[2] = 0;
	TFT_Write_Data_ili9341(0x2B, &uwData[0], 3);  // Page Address Set, we change only the start (2 Parameters)
}

void TFT_Flush_Row_Buff(uint16 numberOfPixel)
{
	uint16 i;
	uint32 val;
    uint16 numberOfWords = numberOfPixel>>1;
    uint16 tx_data;

	TFT_ChipSelect();

    // we write first the address (10 bit) => we change the datawidth to 10 bit
    s_bacon0.B.DL = 9;

    // calculate the command and address value
    tx_data = (0x1 << 8) | 0x2C; // register 0x2C on ILI9341

    // and send the address to ILI9341
    s_bacon0.B.LAST = 0;
    QSPI0_Transfer(tx_data, 1);

    // our values have 2x16 bit  => we change the datawidth to 32 bit
    s_bacon0.B.DL = 31;

    // send dummy values and read the values from ILI9341
    for (i=0; i<numberOfWords-1; i++)
    {
    	val = (Pixel_Buff[i<<1]<<16) | (Pixel_Buff[(i<<1)+1]);
    	if (i==0)
    		QSPI0_Transfer(val, 1);
    	else
    	    QSPI0_Transfer(val, 0);
    }

    s_bacon0.B.LAST = 1;
	val = (Pixel_Buff[i<<1]<<16) | (Pixel_Buff[(i<<1)+1]);
	QSPI0_Transfer(val, 1);

	TFT_ChipUnselect();

    // we terminate the transfer by generate one SCLK without transfer
    // we make sure that the SCLK will be low when we switch to general output
    P20_OMCR.B.PCL11 = 1;
    P20_IOCR8.B.PC11 = 0x10;

    // toggle the SCLK
    P20_OMSR.B.PS11 = 1;
    _nop(); _nop(); _nop(); _nop();
    P20_OMCR.B.PCL11 = 1;

    // set the SCLK back to used alternate output
	P20_IOCR8.B.PC11 = 0x13;

    // Finally set back to 32 bit transfer
    s_bacon0.B.DL = 31;
}

void TFT_ClearScreen(uint16 col)
{
	uint16 i;
	uint32 dataword = ((uint32)col<<16)|col;
    uint16 numberOfWords = TFT_XSIZE*TFT_YSIZE/2+1;
    uint16 tx_data;

    TFT_SetXY (0,0);

	TFT_ChipSelect();

    // we write first the address (10 bit) => we change the datawidth to 10 bit
    s_bacon0.B.DL = 9;

    // calculate the command and address value
    tx_data = (0x1 << 8) | 0x2C; // register 0x2C on ILI9341

    // and send the address to ILI9341
    s_bacon0.B.LAST = 0;
    QSPI0_Transfer(tx_data, 1);

    // our values have 2x16 bit  => we change the datawidth to 32 bit
    s_bacon0.B.DL = 31;

    // send dummy values and read the values from ILI9341
    for (i=0; i<numberOfWords-1; i++)
    {
    	if (i==0)
    		QSPI0_Transfer(dataword, 1);
    	else
    	    QSPI0_Transfer(dataword, 0);
    }

    s_bacon0.B.LAST = 1;
	QSPI0_Transfer(dataword, 1);

	TFT_ChipUnselect();

    // we terminate the transfer by generate one SCLK without transfer
    // we make sure that the SCLK will be low when we switch to general output
    P20_OMCR.B.PCL11 = 1;
    P20_IOCR8.B.PC11 = 0x10;

    // toggle the SCLK
    P20_OMSR.B.PS11 = 1;
    _nop(); _nop(); _nop(); _nop();
    P20_OMCR.B.PCL11 = 1;

    // set the SCLK back to used alternate output
	P20_IOCR8.B.PC11 = 0x13;

    // Finally set back to 32 bit transfer
    s_bacon0.B.DL = 31;
}
