/*====================================================================
* Project:  Board Support Package (BSP)
* Function: Set and determine the frequency the CPU is running at
*
* Copyright HighTec EDV-Systeme GmbH 1982-2008
*====================================================================*/

#include "tc27xc/Ifx_reg.h"
#include <machine/intrinsics.h>
#include <machine/wdtcon.h>
#include <stdio.h>
#include <stdlib.h>

//CCUCON6 is not (yet) defined
#define SCU_CCUCON6  (*(volatile unsigned*)0xF0036080u)


#ifndef DEF_FRQ
#define DEF_FRQ			20000000U	// Ext quartz frequency is 20 MHz
#endif

#define VCOBASE_FREQ	400000000U

// Divider values for 180 MHz from 20 MHz quartz
#define SYS_CFG_PDIV	 2
#define SYS_CFG_NDIV	54
#define SYS_CFG_K1DIV	 3
#define SYS_CFG_K2DIV	 3


// Initialization flag: prevent multiple initialization of PLL_CLC
static unsigned char freq_init = 0;

// Set the frequency of the PLL to the values of N,P,K1,K2 (180 MHz in our case)
// The present configuration of the clock dividers (CCUCONx) is defenitely optimized
// for performance and not for power saving!

void set_cpu_frequency(void)
{
	static Ifx_SCU_PLLCON0 pllcon0;        //Shadow register definition for pllcon0 & 1
	static Ifx_SCU_PLLCON1 pllcon1;

	//Check if PLL has already been configured
	if (freq_init)
		return;

	//Only Core 0 is allowed to change PLL settings
	if ( _mfcr(CPU_CORE_ID) & 0x0f)
		return;

	//Remember that the frequency has been set
	freq_init = 1;

	//Fill the shadow registers with the current values
	pllcon0 = SCU_PLLCON0;
	pllcon1 = SCU_PLLCON1;

	//This function should not be called during power-saving...
	//if (!SCU_PLLSTAT.B.PWDSTAT)
	{
		// Set speed to 180 MHz with 20MHz Crystal in shadow registers
		pllcon0.U = 0;
		pllcon1.U = 0;
		pllcon0.B.NDIV  = SYS_CFG_NDIV - 1;
		pllcon0.B.PDIV  = SYS_CFG_PDIV - 1;
		pllcon1.B.K2DIV = SYS_CFG_K2DIV - 1;
		pllcon1.B.K1DIV = SYS_CFG_K1DIV - 1;
		pllcon0.B.VCOBYP = 1;
		pllcon0.B.CLRFINDIS = 1;
		pllcon0.B.PLLPWD = 1;
		pllcon0.B.RESLD = 1;

		// Select correct OSCVAL for 20 MHz quartz
		unlock_safety_wdtcon();
		SCU_OSCCON.U = 0x0007001C;

		// Set dividers for peripherals:
		//
		// fOSC0 is used as clock source
		// fCAN = fsource (180 MHz)
		// fERAY = fPLLERAY
		// fSTM = fsource (180 MHz)
		// fGTM = fsource/2 (90 MHz)
		// fETH = fPLLERAY/4
		// fASCLINF is stopped
		// fASCLINS is stopped
		// Update request bit is set
		while (SCU_CCUCON1.B.LCK);
		SCU_CCUCON1.U = 0x50012111;

		// fBBB = fsource/2
		// Update request bit is set
		while (SCU_CCUCON2.B.LCK);
		SCU_CCUCON2.U = 0x40000202;

		// Force prescaler mode without PLL and wait
		SCU_PLLCON0.B.VCOBYP = 1;
		while (!SCU_PLLSTAT.B.VCOBYST);

		// fPLL is used as clock source fsource
		// Low Power Divider disabled
		// fBAUD1 = fsource (180 MHz)
		// fBAUD2 = fsource (180 MHz)
		// fSRI = fsource (180 MHz)
		// fSPB = fsource/2 (90 MHz)
		// fFSI  = fSRI / 2 (90 MHz)
		// fFSI2 = fSRI / 2 (90 MHz)
		// fADC = fPLL2
		// Update request bit is set
		SCU_CCUCON0.U =  0x56220111;

		// Write new control values from shadow registers
		SCU_PLLCON1 = pllcon1;
		SCU_PLLCON0 = pllcon0;
		lock_safety_wdtcon();

		// Wait for stable PLL frequency
		while (!SCU_PLLSTAT.B.VCOLOCK);

		// Disable bypass -> use PLL frequency now
		unlock_safety_wdtcon();
		SCU_PLLCON0.B.VCOBYP = 0;
		lock_safety_wdtcon();

	}
}


// Determine the frequency the CPU is running at
// Assumption: All cores are running with the same frequency!

unsigned int get_cpu_frequency(void)
{
	unsigned int frequency;
    unsigned short cpudiv;
	Ifx_SCU_PLLCON0 pllcon0;
	Ifx_SCU_PLLCON1 pllcon1;
	Ifx_SCU_PLLSTAT pllstat;

	//Read out the PLL relevant SFRs
	pllcon0 = SCU_PLLCON0;
	pllcon1 = SCU_PLLCON1;
	pllstat = SCU_PLLSTAT;

	// Read CPU0 divider value
	cpudiv = SCU_CCUCON6 & 0x3f;

	// Look if bypass mode is selected
	if (pllstat.B.VCOBYST)
	{
		unsigned int k_div;

		k_div = pllcon1.B.K1DIV + 1;
		frequency = DEF_FRQ / k_div;
	}
	// Look if divider 'N' is disabled
	else if (pllstat.B.FINDIS)
	{
		unsigned int k_div;

		k_div = pllcon1.B.K2DIV + 1;
		frequency = VCOBASE_FREQ / k_div;
	}
	//Normal PLL mode with fPLL = fosc*n/(k*p)
	else
	{
		unsigned int k_div, n_div, p_div;

		n_div = pllcon0.B.NDIV + 1;
		p_div = pllcon0.B.PDIV + 1;
		k_div = pllcon1.B.K2DIV + 1;

		frequency = DEF_FRQ * n_div / (k_div * p_div);
	}

	//Take CPU0 divider value from CCUCON6 into account
	frequency /= (cpudiv + 1);

	return frequency;
}
