/*
 * Copyright (c) 2001-2004 Swedish Institute of Computer Science.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
 * SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * This file is part of the lwIP TCP/IP stack.
 *
 * Author: Adam Dunkels <adam@sics.se>
 *
 */
#include "tc27xc/Ifx_reg.h"
#include <machine/intrinsics.h>
#include <machine/wdtcon.h>
#include "interrupts.h"
#include "lwip/mem.h"
#include "netif/etharp.h"
#include "ethernetif.h"
#include "IfxEth_phy_Pef7071.h"
#include "Eth_Descr.h"
#include <string.h>
#include "FreeRTOS.h"
#include "queue.h"
#include "semphr.h"
#include "task.h"

// Interrupt priority for ETH interface
#define ETH_INT_PRIO 8

#define netifINTERFACE_TASK_STACK_SIZE		( 350 )
#define netifINTERFACE_TASK_PRIORITY		( 5 )


// Defines from Infineons iLLD
#define IFX_ETH_STATUS_RS_OFF (17)
#define IFX_ETH_STATUS_RU_MSK (0x1)
#define IFX_ETH_STATUS_RU_OFF (7)
#define IFX_ETH_STATUS_TU_MSK (0x1)
#define IFX_ETH_STATUS_TU_OFF (2)
#define IFX_ETH_STATUS_UNF_MSK (0x1)
#define IFX_ETH_STATUS_UNF_OFF (5)


/* Network interface name */
#define IFNAME0 't'
#define IFNAME1 'c'
#define IfxCpu_getCoreId() (_mfcr(CPU_CORE_ID) & 0x0f)

/* Definitions needed for ETH device */
static unsigned int ETH_error;
static Ifx_ETH_STATUS ETH_status;
static unsigned int ETH_rxCount;
static unsigned int ETH_txCount;
static IfxEth_RxDescrList ETH_rxDescr;
static IfxEth_TxDescrList ETH_txDescr;
static unsigned char ETH_rxBuffer[ETH_RX_BUFFERS][ETH_RTX_BUFFER_SIZE];
static unsigned char ETH_txBuffer[ETH_TX_BUFFERS][ETH_RTX_BUFFER_SIZE];
static IfxEth_RxDescr *ETH_pRxDescr;
static IfxEth_TxDescr *ETH_pTxDescr;
static xSemaphoreHandle ETH_TxSem;
static xQueueHandle ETH_RxQueue;
static struct pbuf *s_p;

extern const unsigned char MAC_ADDRESS[6];

void ETH_IRQ_Handler();

#define P11_PCSR (*(volatile unsigned*)0xF003B164)

void ETH_startReceiver()
{
    // Enable receiver and RX DMA
    ETH_OPERATION_MODE.B.SR    = 1;
    ETH_MAC_CONFIGURATION.B.RE = 1;
    ETH_RECEIVE_POLL_DEMAND.U  = 1;
}


void ETH_startTransmitter()
{
    ETH_MAC_CONFIGURATION.B.TE = 1;
    ETH_OPERATION_MODE.B.ST    = 1;
    ETH_TRANSMIT_POLL_DEMAND.U = 1;
}


void ETH_stopTransmitter()
{
    ETH_TRANSMIT_POLL_DEMAND.U = 0;
    ETH_OPERATION_MODE.B.ST    = 0;
    ETH_MAC_CONFIGURATION.B.TE = 0;
}

void ETH_wakeupReceiver()
{
    ETH_status.U = ETH_STATUS.U;

    // check if receiver suspended
    if (ETH_status.U & (4 << IFX_ETH_STATUS_RS_OFF))
    {
        if (ETH_status.B.RU)
        {
            ETH_STATUS.U = (IFX_ETH_STATUS_RU_MSK << IFX_ETH_STATUS_RU_OFF);
        }

        ETH_startReceiver();
    }
}


void ETH_wakeupTransmitter()
{
    ETH_status.U = ETH_STATUS.U;

    // check if suspended
    if (ETH_status.U & 0x00600000)
    {
        if (ETH_status.B.TU)
        {
            // clear transmit unavailable and underflow flags
            ETH_STATUS.U = (IFX_ETH_STATUS_TU_MSK << IFX_ETH_STATUS_TU_OFF) |
                           (IFX_ETH_STATUS_UNF_MSK << IFX_ETH_STATUS_UNF_OFF);
        }

        ETH_startTransmitter();
    }
}

void ETH_initReceiveDescriptors()
{
    // Use enhanced descriptors, because ETH is generally configured to precision time protocol.
    int i;
    IfxEth_RxDescr *descr = ETH_rxDescr.items;

    ETH_pRxDescr = descr;

    // init descriptor chained mode
    for (i = 0; i < ETH_RX_BUFFERS; i++)
    {
        descr->RDES0.U      = 0;
        descr->RDES0.A.OWN  = 1;
        descr->RDES1.U      = 0;
        descr->RDES1.A.RCH  = 1;
        descr->RDES1.A.RBS1 = (ETH_RTX_BUFFER_SIZE);
        descr->RDES2.U = (uint32)IFXCPU_GLB_ADDR_DSPR(IfxCpu_getCoreId(), &ETH_rxBuffer[i][0]);

        // with RCH set, link to next descriptor address
        descr->RDES3.U = (uint32)IFXCPU_GLB_ADDR_DSPR(IfxCpu_getCoreId(), &(descr[1]));
        descr = &descr[1];
    }

    // correction for last descriptor
    {
        descr = &descr[-1];

        // indicate end of ring
        descr->RDES1.A.RER = 1;

        // with RCH set, link to first descriptor address
        ETH_pRxDescr  = ETH_rxDescr.items;
        descr->RDES3.U = (uint32)IFXCPU_GLB_ADDR_DSPR(IfxCpu_getCoreId(), ETH_pRxDescr);
    }

    ETH_rxCount = 0;

    // write descriptor list base address
    ETH_RECEIVE_DESCRIPTOR_LIST_ADDRESS.U = (uint32)IFXCPU_GLB_ADDR_DSPR(IfxCpu_getCoreId(),(unsigned)ETH_rxDescr.items);
}


void ETH_initTransmitDescriptors()
{
    int i;
    IfxEth_TxDescr *descr = ETH_txDescr.items;

    ETH_pTxDescr = descr;

    // Initialize chained descriptor mode
    for (i = 0; i < ETH_TX_BUFFERS; i++)
    {
        descr->TDES0.U     = 0;
        descr->TDES0.A.IC  = 1;
        descr->TDES0.A.FS  = 1;
        descr->TDES0.A.LS  = 1;
        descr->TDES0.A.TCH = 1;
        descr->TDES2.U = (uint32)IFXCPU_GLB_ADDR_DSPR(IfxCpu_getCoreId(), &ETH_txBuffer[i][0]);

        // with TCH set, TDES3 points to next descriptor
        descr->TDES3.U = (uint32)IFXCPU_GLB_ADDR_DSPR(IfxCpu_getCoreId(), &descr[1]);
        descr = &descr[1];
    }

    // correction for last descriptor
    {
        descr = &descr[-1];

        // indicate end of ring
        descr->TDES0.A.TER = 1;

        // with TCH set, TDES3 points to the first descriptor
        ETH_pTxDescr  = ETH_txDescr.items;
        descr->TDES3.U = (uint32)IFXCPU_GLB_ADDR_DSPR(IfxCpu_getCoreId(), ETH_pTxDescr);
    }

    ETH_txCount = 0;

    // write descriptor list base address
    ETH_TRANSMIT_DESCRIPTOR_LIST_ADDRESS.U = (uint32)IFXCPU_GLB_ADDR_DSPR(IfxCpu_getCoreId(),(unsigned)ETH_txDescr.items);
}

void ETH_freeReceiveBuffer()
{
    IfxEth_RxDescr *descr = ETH_pRxDescr;

    descr->RDES0.A.OWN = 1;
    ETH_pRxDescr = (IfxEth_RxDescr *)(ETH_pRxDescr->RDES3.U);
}

void *ETH_getReceiveBuffer(uint16 *framelength)
{
    void *result = NULL;
    IfxEth_RxDescr *descr;

    if (ETH_pRxDescr->RDES0.A.OWN == 0)
    {
        ETH_rxCount++;
        descr  = ETH_pRxDescr;
        *framelength = descr->RDES0.A.FL;
        result = (void *)(descr->RDES2.U);
    }

    ETH_wakeupReceiver();

    return result;
}


void *ETH_getTransmitBuffer()
{
    void *buffer = NULL;
    IfxEth_TxDescr *descr  = ETH_pTxDescr;

    // check descriptor / buffer is free.
    if (descr->TDES0.A.OWN == 0)
    {
        buffer = ((void *)descr->TDES2.U);
    }

    return buffer;
}


void ETH_Init()
{
	//Shadow registers
	Ifx_ETH_BUS_MODE busMode;
    Ifx_ETH_MAC_CONFIGURATION ethMacCfg;


	//Enable ETH module by enabling its clock
	unlock_wdtcon();
    ETH_CLC.U = 0;
	lock_wdtcon();

    // Configure ETH pins according to page 5046 of UM //

	//Configure Pin P11.12 for ETHRXCLKA
	P11_IOCR12.B.PC12=0x00;
	ETH_GPCTL.B.ALTI1=0x00;

	//Configure Pin 11.11 for ETHRXDVA
	P11_IOCR8.B.PC11 = 0x00;
	ETH_GPCTL.B.ALTI4 = 0x00;

	//Configure Pin 11.9 for ETHRXD1A
	P11_IOCR8.B.PC9 = 0x00;
	ETH_GPCTL.B.ALTI7 = 0x00;

	//Configure Pin 11.10 for ETHRXD0A
	P11_IOCR8.B.PC10 = 0x00;
	ETH_GPCTL.B.ALTI6 = 0x00;

	//Configure Pin 11.6 for ETHTXEN
	P11_IOCR4.B.PC6 = 0x16;

	//Configure Pin 11.2 for ETHTXD1
	P11_IOCR0.B.PC2 = 0x16;

    //Configure Pin 11.3 for ETHTXD0
	P11_IOCR0.B.PC3 = 0x16;

	//Configure Pin 21.3 for ETHMDIOD
	P21_IOCR0.B.PC3 = 0x00;
	ETH_GPCTL.B.ALTI0 = 0x03;

    //Configure Pin 21.2 for ETHMDC
	P21_IOCR0.B.PC2 = 0x15;

	//Configure Pin 20.9 for SCU input
	P20_IOCR8.B.PC9 = 0x00;

	unlock_safety_wdtcon();
	P11_PCSR = (1<<6) | (1<<2) | (1<<3);
	lock_safety_wdtcon();
	unlock_wdtcon();
	P11_PDR0.U = 0x11111111;
	P11_PDR1.B.PD12 = 0;
	P21_PDR0.B.PD2 = 1;
	lock_wdtcon();

	//Configure PEF7071
	IfxEth_Phy_Pef7071_init();

	//Set Phy Interface Mode rmii
	ETH_GPCTL.B.EPR = 1;

	//Do software reset of ETH module and wait for its completion
	ETH_BUS_MODE.B.SWR = 1;
    while (ETH_BUS_MODE.B.SWR);

    //Configure bus mode
	busMode.U = ETH_BUS_MODE.U;
	busMode.B.DSL = 0;
	busMode.B.ATDS = 0;
	busMode.B.DA = 0;
	ETH_BUS_MODE.U = busMode.U;

    ethMacCfg.U = ETH_MAC_CONFIGURATION.U;
    ethMacCfg.B.PRELEN = 0;  // 7 bytes preamble
    //  ethMacCfg.B.RE = 0;  // disable receiver
    //  ethMacCfg.B.TE = 0;  // disable transmitter
    ethMacCfg.B.DC = 0;      // Deferral Check
    //  ethMacCfg.B.BL = 0;  // Backoff Limit
    ethMacCfg.B.ACS = 1;     // Automatic Pad/CRC stripping
    //  ethMacCfg.B.DR = 0;  // Disable Retry
    ethMacCfg.B.IPC = 0;     // checksum offload
    ethMacCfg.B.DM = 1;      // Duplex Mode: 0=Half Duplex, 1=Full duplex
    ethMacCfg.B.LM = 0;      // Loopback Mode
    //  ethMacCfg.B.DO = 0;  // Disable Receive Own
    ethMacCfg.B.FES = 1;     // Speed 0->10mbps 1->100mbps
    ethMacCfg.B.PS = 1;      // port select 10/100mbps
    ethMacCfg.B.IFG = 0;     // Inter Frame Gap - gap between frames = 96 bit times
    ethMacCfg.B.JE = 0;      // No jumbo frames
    ethMacCfg.B.JD = 0;      // Jabber Disable - cut of transmission after 2,048 data bytes.
    ethMacCfg.B.WD = 0;      // Watchdog Disable - cut off frame after 2,048 bytes.
    ethMacCfg.B.CST = 1;     // CRC stripping - last four bytes are stripped and dropped to application.
    ethMacCfg.B.TWOKPE = 0;  // 2K Packets Enable - with JE=0
    ETH_MAC_CONFIGURATION.U = ethMacCfg.U;

    ETH_MAC_ADDRESS_G00_HIGH.U = ((uint32)MAC_ADDRESS[4] << 0)
                               | ((uint32)MAC_ADDRESS[5] << 8)
                               | 0x80000000U;

    ETH_MAC_ADDRESS_G00_LOW.U = ((uint32)MAC_ADDRESS[0] << 0)
                              | ((uint32)MAC_ADDRESS[1] << 8)
                              | ((uint32)MAC_ADDRESS[2] << 16)
                              | ((uint32)MAC_ADDRESS[3] << 24);

    // Setup MMC
    ETH_MMC_CONTROL.B.CNTFREEZ = 1;         // Disable MMC counters - counters reset

    // Setup GMAC
    ETH_STATUS.U           = 0x0001e7ff;    // Reset all interrupt flag(s)
    ETH_MAC_FRAME_FILTER.U = 0x00000010;    // Hash Unicast

    ETH_INTERRUPT_ENABLE.U = 0x000180c5;    // Enable TX & RX & Buffer Underrun interrupts

	InterruptInstall(SRC_ID_ETH, ETH_IRQ_Handler, ETH_INT_PRIO, 0);

	//ATEN: Guess this is not needed anymore, or?
	//config->phyInit();          // init PHY (100Mbit, full duplex with RMII)

	ETH_error    = 0;
    ETH_status.U = 0;
    ETH_rxCount  = 0;
    ETH_txCount  = 0;

    //Configure the RX Queue for incoming messages
    ETH_RxQueue= xQueueCreate(ETH_RX_BUFFERS, 4);

    ETH_stopTransmitter();

    ETH_initReceiveDescriptors();
    ETH_initTransmitDescriptors();
}

void ETH_Start()
{
    ETH_startTransmitter();
    ETH_startReceiver();
}

unsigned short int ETH_int_count=0;

void ETH_IRQ_Handler()
{
    portBASE_TYPE taskWoken = pdFALSE;
    unsigned int val;
    uint16 reason;
    uint8 mustYield = 0;

    ETH_int_count++;

	reason = ETH_STATUS.U;

    while (reason & ((1<<6)|(1<<0)))
    {

		//Check for RX interrupt
		if (reason & (1<<6) )
		{
			if (ETH_pRxDescr->RDES0.A.OWN == 0)
			{
				val = (unsigned)ETH_pRxDescr;
				xQueueSendFromISR(ETH_RxQueue, &val, &taskWoken);
				mustYield |= taskWoken;
			}
			ETH_STATUS.U = 1<<6;
		}

		//Check for RX buffer unavailable interrupt
		//Happens usually during debugging
		if (reason & (1<<7) )
		{
			ETH_startReceiver();
			ETH_STATUS.U = 1<<7;
		}

		//Check for TX interrupt
		if (reason & (1<<0))
		{
			//Since TX interrupt is obviously not working with small frames use polling
			//in low_level_input function in the rare case that all TX buffers are in use by the DMA
			//xSemaphoreGiveFromISR(ETH_TxSem, &taskWoken);
			//mustYield |= taskWoken;
			ETH_STATUS.U = 1<<0;
		}

		//Check for TX buffer unavailable interrupt
		//Happens usually during debugging
		if (reason & (1<<2))
		{
			ETH_startTransmitter();
			ETH_STATUS.U = 1<<2;
		}

		reason = ETH_STATUS.U;
    }

	if (mustYield)
	{
	   taskYIELD();
	}

    //Clear all pending interrupts
    ETH_STATUS.U = 0x0001e7ff;
}

void ETH_sendTransmitBuffer(uint16 len)
{
    IfxEth_TxDescr *descr = ETH_pTxDescr;

    descr->TDES1.U     = len;      // with TCH set, TBS1 is used for buffer size
    descr->TDES0.A.OWN = 1;        // release to DMA

    ETH_wakeupTransmitter();

    //Select the next descriptor
    ETH_pTxDescr = (IfxEth_TxDescr*)ETH_pTxDescr->TDES3.U;

    ETH_txCount++;
}


static struct pbuf * low_level_input(struct netif *netif)
{
	  struct pbuf *q;
	  uint16 length = 0;
	  uint16 framelength = 0;

	  s_p = NULL;

      if (ETH_pRxDescr->RDES0.A.OWN == 0)
      {
    	  // We expect that a packet is available
		  unsigned char *pRxBuf = ETH_getReceiveBuffer(&framelength);

		  // We allocate a pbuf chain of pbufs from the pool.
		  s_p = pbuf_alloc(PBUF_RAW, framelength, PBUF_POOL);

		  if (s_p == NULL)
			for (;;);

		  // Copy received frame from Ethernet driver buffer to stack buffer pbufs chain

		  for(q = s_p; q != NULL; q = q->next)
		  {
		     memcpy(q->payload, &pRxBuf[length], q->len);
			 length = length + q->len;
		  }

	      ETH_freeReceiveBuffer();
      }

	  return s_p;
}

err_t ethernetif_input(struct netif *netif);

/**
 * In this function, the hardware should be initialized.
 * Called from ethernetif_init().
 *
 * @param netif the already initialized lwip network interface structure
 *        for this ethernetif
 */
static void low_level_init(struct netif *netif)
{
  /* set MAC hardware address length */
  netif->hwaddr_len = ETHARP_HWADDR_LEN;

  /* set MAC hardware address within netif struct*/
  netif->hwaddr[0] =  MAC_ADDRESS[0];
  netif->hwaddr[1] =  MAC_ADDRESS[1];
  netif->hwaddr[2] =  MAC_ADDRESS[2];
  netif->hwaddr[3] =  MAC_ADDRESS[3];
  netif->hwaddr[4] =  MAC_ADDRESS[4];
  netif->hwaddr[5] =  MAC_ADDRESS[5];
  
  /* maximum transfer unit */
  netif->mtu = 1500;

  /* device capabilities */
  /* don't set NETIF_FLAG_ETHARP if this device is not an ethernet one */
  netif->flags = NETIF_FLAG_BROADCAST | NETIF_FLAG_ETHARP | NETIF_FLAG_LINK_UP;

  ETH_TxSem = xSemaphoreCreateCounting(ETH_TX_BUFFERS, ETH_TX_BUFFERS);

  xTaskCreate(ethernetif_input, "Eth_if", netifINTERFACE_TASK_STACK_SIZE, netif, netifINTERFACE_TASK_PRIORITY, NULL);
  
  // Enable MAC and DMA transmission and reception
  ETH_Start();

}

/**
 * This function should do the actual transmission of the packet. The packet is
 * contained in the pbuf that is passed to the function. This pbuf
 * might be chained.
 *
 * @param netif the lwip network interface structure for this ethernetif
 * @param p the MAC packet to send (e.g. IP packet including MAC addresses and type)
 * @return ERR_OK if the packet could be sent
 *         an err_t value if the packet couldn't be sent
 *
 * @note Returning ERR_MEM here if a DMA queue of your MAC is full can lead to
 *       strange results. You might consider waiting for space in the DMA queue
 *       to become availale since the stack doesn't retry to send a packet
 *       dropped because of memory failure (except for the TCP timers).
 */


static err_t low_level_output(struct netif *netif, struct pbuf *p)
{
  struct pbuf *q;
  uint16 framelength = 0;
  unsigned char *pTxBuf = NULL;

  //Wait for a free TX buffer
  //xSemaphoreTake(ETH_TxSem , 0xffffffff);

  //Get a pointer to the free TX buffer - use busy waiting in case non is free
  while (pTxBuf == NULL)
  {
	  pTxBuf = (unsigned char*)ETH_getTransmitBuffer();
  }
  
  // Copy frame from pbufs to ETH buffer
  for(q = p; q != NULL; q = q->next) 
  {
     memcpy(&pTxBuf[framelength], q->payload, q->len);
	 framelength = framelength + q->len;
  }
  
  //And transmit the frame
  ETH_sendTransmitBuffer(framelength);

  return ERR_OK;
}


/**
 * Should allocate a pbuf and transfer the bytes of the incoming
 * packet from the interface into the pbuf.
 *
 * @param netif the lwip network interface structure for this ethernetif
 * @return a pbuf filled with the received packet (including MAC header)
 *         NULL on memory error
 */


/**
 * This function should be called when a packet is ready to be read
 * from the interface. It uses the function low_level_input() that
 * should handle the actual reception of bytes from the network
 * interface. Then the type of the received packet is determined and
 * the appropriate input function is called.
 *
 * @param netif the lwip network interface structure for this ethernetif
 */
err_t ethernetif_input(struct netif *netif)
{
	unsigned int val;

	for( ;; )
	{
		if (xQueueReceive(ETH_RxQueue, &val, 0xffff))
		{
			do
			{
			    s_p = low_level_input( netif );
			    if (s_p)
			    {
					if (ERR_OK != netif->input( s_p, netif))
					{
						pbuf_free(s_p);
						s_p = NULL;
					}
			    }
			}
			while (s_p);
		}
	}

/*  err_t err;
  struct pbuf *p;

  // move received packet into a new pbuf
  p = low_level_input(netif);

  // no packet could be read, silently ignore this
  if (p == NULL) return ERR_MEM;

  // entry point to the LwIP stack
  err = netif->input(p, netif);
  
  if (err != ERR_OK)
  {
    LWIP_DEBUGF(NETIF_DEBUG, ("ethernetif_input: IP input error\n"));
    pbuf_free(p);
    p = NULL;
  }
  return err;*/
}

/**
 * Should be called at the beginning of the program to set up the
 * network interface. It calls the function low_level_init() to do the
 * actual setup of the hardware.
 *
 * This function should be passed as a parameter to netif_add().
 *
 * @param netif the lwip network interface structure for this ethernetif
 * @return ERR_OK if the loopif is initialized
 *         ERR_MEM if private data couldn't be allocated
 *         any other err_t on error
 */
err_t ethernetif_init(struct netif *netif)
{
  LWIP_ASSERT("netif != NULL", (netif != NULL));
  
  netif->name[0] = IFNAME0;
  netif->name[1] = IFNAME1;
  /* We directly use etharp_output() here to save a function call.
   * You can instead declare your own function an call etharp_output()
   * from it if you have to do some checks before sending (e.g. if link
   * is available...) */
  netif->output = etharp_output;
  netif->linkoutput = low_level_output;

  /* initialize the hardware */
  low_level_init(netif);

  return ERR_OK;
}



