/*
    FreeRTOS V7.1.0 - Copyright (C) 2011 Real Time Engineers Ltd.

    FreeRTOS is free software; you can redistribute it and/or modify it under
    the terms of the GNU General Public License (version 2) as published by the
    Free Software Foundation AND MODIFIED BY the FreeRTOS exception.
    >>>NOTE<<< The modification to the GPL is included to allow you to
    distribute a combined work that includes FreeRTOS without being obliged to
    provide the source code for proprietary components outside of the FreeRTOS
    kernel.  FreeRTOS is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
    more details. You should have received a copy of the GNU General Public
    License and the FreeRTOS license exception along with FreeRTOS; if not it
    can be viewed here: http://www.freertos.org/a00114.html and also obtained
    by writing to Richard Barry, contact details for whom are available on the
    FreeRTOS WEB site.
*/

/*
 * This is the startup file for Core #0 for the FreeRTOS Multicore port
 * done by the Virtual Vehicle Research and Competence Center in Graz, Austria
 *
 * The demonstration projects shows how to start up all 3 cores and running
 * their own instance of FreeRTOS on it. It is intended to be compiled
 * using the Free Entry Toolchain provided by Infineon and run on a Eval board
 * "Application Kit TC2X7 V1.0", equipped with a TC277 Step C controller.
 * You'll need to compile the project with the "-fdollars-in-identifiers" option!
 *
 * As part of the demonstration, a mini web server is started on the port, using
 * the MAC address CAFE000001 and the static IP address 192.168.0.100 (main1.c).
 * Furthermore some tasks for demonstrating the inter-core communication are
 * started and a heart-beat task flashing LED 13.0 is running in the background.
 *
 * Because of the local memory map vs. global memory map of the AURIX MC
 * devices, quite some memory is wasted because the compiler is not aware
 * of the local memory at 0xD000000. Thus all global variables are available
 * on all three cores, but typically they are only used in one core, on the
 * other cores the same memory range is 'wasted'.
 * A work around is using the heap and stack memory, because every core
 * maintains its own heap and every task has its own stack...
 *
 * For inter-core communication a Shared Memory Fifo has been implemented.
 * See sharedmem.c of how to use it!
 *
 * Have fun working with FreeRTOS on the Aurix controller family!
 *
 * Copyright (c) 2015 by Virtual Vehicle Research and Competence Center
 * Author: Allan Tengg (allan.tengg@v2c2.at)
 *
 */


#include "tc27xc/Ifx_reg.h"
#include <machine/intrinsics.h>
#include <machine/wdtcon.h>
#include "system_tc2x.h"
#include "interrupts.h"
#include "cpufreq.h"
#include "led.h"
#include "debug.h"
#include "timer.h"
#include <stdlib.h>
#include <string.h>
#include "led.h"
#include "sharedmem.h"
#include "TLF35584.h"

extern unsigned int _start;
extern unsigned int __ISTACK;

extern void main0();
extern void main1();
extern void main2();

typedef void (*FP)(void);
const FP main_routines[3] = {main0, main1, main2};
Ifx_CPU* cpus[3] = {&MODULE_CPU0, &MODULE_CPU1, &MODULE_CPU2};


void startCore(unsigned short cpunr, unsigned int programCounter)
{
	if (cpunr > 2)
	  return;

	cpus[cpunr]->PC.B.PC = programCounter >> 1;

	unlock_safety_wdtcon();
    MODULE_SCU.PMCSR[cpunr].B.REQSLP = 0;
    lock_safety_wdtcon();

    cpus[cpunr]->DBGSR.B.HALT = 2;  //reset the HALT bit
}

//*******************************************************************************************
//** main()
//**
//** This is the common entry point for all CPU cores (0,1,2). The startup code has prepared
//** at this point the stack and context areas of every core identically.
//**
//** As the starting core, core #0 is responsible to perform all crucial initialization
//** routines before starting the other cores. This includes disabling the watchdog
//** within the TLF35584 safety power supply chip and setting the desired PLL frequencies.
//**
//** After that it is safe to turn on the other cores with the C initialization code.
//** They will too jump to main routine, initialize their interrupt- and trap-tables
//** and branch to the corresponding 'main0'...'main2' routine where individual
//** initialization steps may be performed.

int main( void )
{
	unsigned short cpu_id = _mfcr(CPU_CORE_ID) & 0x0f;
	unsigned int intstack = &__ISTACK;

	if (cpu_id == 0)
	{
		//First of all disable the TLF35584 safety chip watchdog to prevent reset in 600 ms
		Disable_TLF35584();

		// Initialize LEDs
		InitLED();
		ClearLED(1);
		ClearLED(2);
		ClearLED(3);

		// Configure the PLL to 180 MHz
		set_cpu_frequency();

		// Start the other cores now
		startCore(1, (unsigned int)&_start);
		startCore(2, (unsigned int)&_start);
	}

	// The interrupt stack pointer is not set correctly on core 1,2 with the
	// default startup code
	unlock_wdtcon();
    __mtcr(CPU_ISP, intstack);
    _isync(); _dsync();
	lock_wdtcon();

	// Setup interrupt handlers
	InterruptInit();

	//Initialize the shared memory communication
	SharedMemInit();

	//All basic initialization has been done ->
	//Jump to the main routine of the correspondig core
	main_routines[cpu_id]();

	return 0;
}

void vApplicationTickHook( void )
{
}

void vApplicationIdleHook( void )
{
	// It will be called on each iteration of the idle task.
	// It is essential that code never attempts to block in any way
	_nop();
}

